@@ -35,7 +35,7 @@ extern int WIN32_SSL_password_prompt(tcn_pass_cb_t *data);
 #define ASN1_OID      0x06
 #define ASN1_STRING   0x86
 static int ssl_verify_OCSP(int ok, X509_STORE_CTX *ctx);
-static int ssl_ocsp_request(X509 *cert, X509 *issuer);
+static int ssl_ocsp_request(X509 *cert, X509 *issuer, X509_STORE_CTX *ctx);
 #endif
 
 /*  _________________________________________________________________
@@ -519,21 +519,21 @@ static int ssl_verify_OCSP(int ok, X509_STORE_CTX *ctx)
     }
 
     /* if we can't get the issuer, we cannot perform OCSP verification */
-    if (X509_STORE_CTX_get1_issuer(&issuer, ctx, cert) == 1 ) {
-        r = ssl_ocsp_request(cert, issuer);
-        if (r == OCSP_STATUS_REVOKED) {
+    issuer = X509_STORE_CTX_get0_current_issuer(ctx);
+    if (issuer != NULL) {
+        r = ssl_ocsp_request(cert, issuer, ctx);
+        switch (r) {
+        case OCSP_STATUS_OK:
+            X509_STORE_CTX_set_error(ctx, X509_V_OK);
+            break;
+        case OCSP_STATUS_REVOKED:
             /* we set the error if we know that it is revoked */
             X509_STORE_CTX_set_error(ctx, X509_V_ERR_CERT_REVOKED);
+            break;
+        case OCSP_STATUS_UNKNOWN:
+            /* ssl_ocsp_request() sets the error correctly already. */
+            break;
         }
-        else {
-            /* else we return unknown */
-            r = OCSP_STATUS_UNKNOWN;
-        }
-        X509_free(issuer); /* It appears that we  should free issuer since
-                            * X509_STORE_CTX_get1_issuer() calls X509_OBJECT_up_ref_count()
-                            * on the issuer object (unline X509_STORE_CTX_get_current_cert()
-                            * that just returns the pointer
-                            */
     }
     return r;
 }
@@ -1009,11 +1009,12 @@ end:
 /* Process the OCSP_RESPONSE and returns the corresponding
    answert according to the status.
 */
-static int process_ocsp_response(OCSP_RESPONSE *ocsp_resp)
+static int process_ocsp_response(OCSP_RESPONSE *ocsp_resp, X509 *cert, X509 *issuer)
 {
     int r, o = V_OCSP_CERTSTATUS_UNKNOWN, i;
     OCSP_BASICRESP *bs;
     OCSP_SINGLERESP *ss;
+    OCSP_CERTID *certid;
 
     r = OCSP_response_status(ocsp_resp);
 
@@ -1023,7 +1024,13 @@ static int process_ocsp_response(OCSP_RESPONSE *ocsp_resp)
     }
     bs = OCSP_response_get1_basic(ocsp_resp);
 
-    ss = OCSP_resp_get0(bs,0); /* we know we have only 1 request */
+    certid = OCSP_cert_to_id(NULL, cert, issuer);
+    if (certid == NULL) {
+        OCSP_RESPONSE_free(ocsp_resp);
+        return OCSP_STATUS_UNKNOWN;
+    }
+    ss = OCSP_resp_get0(bs, OCSP_resp_find(bs, certid, -1)); /* find by serial number and get the matching response */
+
 
     i = OCSP_single_get0_status(ss, NULL, NULL, NULL, NULL);
     if (i == V_OCSP_CERTSTATUS_GOOD)
@@ -1034,11 +1041,12 @@ static int process_ocsp_response(OCSP_RESPONSE *ocsp_resp)
         o = OCSP_STATUS_UNKNOWN;
 
     /* we clean up */
+    OCSP_CERTID_free(certid);
     OCSP_RESPONSE_free(ocsp_resp);
     return o;
 }
 
-static int ssl_ocsp_request(X509 *cert, X509 *issuer)
+static int ssl_ocsp_request(X509 *cert, X509 *issuer, X509_STORE_CTX *ctx)
 {
     char **ocsp_urls = NULL;
     int nid;
@@ -1061,13 +1069,20 @@ static int ssl_ocsp_request(X509 *cert, X509 *issuer)
        the ocsp status. Otherwise, return OCSP_STATUS_UNKNOWN */
     if (ocsp_urls != NULL) {
         OCSP_RESPONSE *resp;
+        int rv = OCSP_STATUS_UNKNOWN;
         /* for the time being just check for the fist response .. a better
            approach is to iterate for all the possible ocsp urls */
         resp = get_ocsp_response(cert, issuer, ocsp_urls[0]);
+        if (resp != NULL) {
+            rv = process_ocsp_response(resp, cert, issuer);
+        } else {
+            /* correct error code for application errors? */
+            X509_STORE_CTX_set_error(ctx, X509_V_ERR_APPLICATION_VERIFICATION);
+        }
 
         if (resp != NULL) {
             apr_pool_destroy(p);
-            return process_ocsp_response(resp);
+            return rv;
         }
     }
     apr_pool_destroy(p);
